/*
    e2floppy.cc

    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/



#include <stdlib.h>
#include "misc.h"
#include "e2floppy.h"

// supported formats used by Ben Williams FLEX Emulator:

struct s_formats formats[] = {
	{ 89600L,  35, 10 },
	{ 179200L, 35, 20 },
	{ 358400L, 35, 40 },
	{ 102400L, 40, 10 },
	{ 204800L, 40, 20 },
	{ 409600L, 40, 40 },
	{ 819200L, 80, 40 },
	{ 0L,      0, 0 }
};

void E2floppy::header_for_own_format(
	s_floppy *pfloppy,
	s_flex_header *pheader,
	int wp)
{
	pfloppy->ready         = 1;
	pfloppy->type          = OWN_FORMAT;
	pfloppy->offset        = sizeof(struct s_flex_header);
	pfloppy->write_protect =
		(wp | pheader->write_protect) ? 0x40 : 0;
	pfloppy->max_sector    = pheader->sectors;
	pfloppy->max_sector0   = pheader->sectors0;
	pfloppy->max_track     = pheader->tracks - 1;
	pfloppy->byte_p_sector =  128 << pheader->sizecode;
	pfloppy->byte_p_track0 =
		pheader->sides0 * pheader->sectors0 * floppy->byte_p_sector;
	pfloppy->byte_p_track  =
		pheader->sides  * pheader->sectors  * floppy->byte_p_sector;
} // header_for_own_format

void E2floppy::header_for_ben_format(s_floppy *pfloppy, s_formats *pformat, int wp) {
	pfloppy->ready         = 1;
	pfloppy->type          = BEN_FORMAT;
	pfloppy->offset        = 0;
	pfloppy->write_protect = wp;
	pfloppy->max_sector    = pformat->sectors >> 1;
	pfloppy->max_sector0   = pformat->sectors >> 1;
	pfloppy->max_track     = pformat->tracks - 1;
	pfloppy->byte_p_sector = 256;
	pfloppy->byte_p_track0 = pformat->sectors << 8; 
	pfloppy->byte_p_track  = pformat->sectors << 8; 
} // header_for_ben_format


E2floppy::E2floppy(Mc6809* x_cpu, char *drive[])
{
	int i;

	cpu = x_cpu;
	for (i = 0; i < 4; i++) {
		pfs = &floppy[i];	// set pfs to selected drive
		pfs->type  = NO_FORMAT;
		pfs->ppath = drive[i];
		pfs->track = 1;	// position to a track != 0  !!!
		pfs->ready = 0;

		if(!init_disk_image(pfs)) {
#ifdef NAFS
			init_nafs_drive(pfs);
#endif // NAFS
		} // if
	}  // for
	selected = 4;			// deselect all drives
	pfs = &floppy[selected];	// set pfs to selected drive
	pfs->ready = 0; 		// drive number 4 is always not ready
} // E2floppy


char *E2floppy::open_mode(char *path)
{
	int  wp;	// write protect flag
	char *mode;
	
	wp = access(path, W_OK);
#ifdef sun
	mode = wp ? "r" : "r+";
#else
	mode = wp ? "rb" : "rb+";
#endif	
	return mode;
} // open_mode


int E2floppy::init_disk_image(struct s_floppy *pfloppy)
{
	int			wp;	// write protect flag
	struct stat		sbuf;
	struct s_flex_header	header;
	struct s_formats	*pformat;

	if(stat(pfloppy->ppath, &sbuf))
		return 0;
	if (S_ISREG(sbuf.st_mode)) {
		wp = access(pfloppy->ppath, W_OK);
		if ((pfloppy->fp = fopen(pfloppy->ppath,
			open_mode(pfloppy->ppath))) == NULL)
			return 0;
		if ((fread((void *)&header, sizeof(struct s_flex_header), 1,
		   pfloppy->fp) == 1) && 
		   (header.magic_number == MAGIC_NUMBER)) {
		   	header_for_own_format(pfloppy, &header, wp);
			return 1;	// it's a disk image with own format 
		} // if
		pformat = formats;
		while (pformat->size) {
			if (pformat->size == sbuf.st_size) {
				header_for_ben_format(pfloppy, pformat, wp);
				return 1; // it's ben's format 
			}
			pformat++; 
		} // while
	}  // if
	return 0;
} // init_disk_image


E2floppy::~E2floppy(void)
{
	int i;

	for (i = 0; i < 4; i++) {
		if (floppy[i].fp != NULL)
			fclose(floppy[i].fp);
#ifdef NAFS
		umount_nafs(i);
#endif
	}
} // ~E2floppy


void E2floppy::reset()
{
	drisel = 0;
	Wd1793::reset();
}


Byte E2floppy::readstatus(Word offset)
{
	status = 0xff;	// unused is logical high
	if (!side)
		status &= 0xfd;
	if (!irq)
		status &= 0xbf;
	if (!drq)
		status &= 0x7f;
	return status;
} // readstatus


void E2floppy::writedrisel(Word offset, Byte val)
{
	drisel = val;
	side = (drisel & 0x10) ? 1 : 0;
	switch (drisel & 0x0f) {
		case 0x01 : selected = 0;
			    break;
		case 0x02 : selected = 1;
			    break;
		case 0x04 : selected = 2;
			    break;
		case 0x08 : selected = 3;
			    break;
		default   : selected = 4;
	};
	pfs->track = tr;
	pfs = &floppy[selected];
	tr = pfs->track;
} // writedrisel


int E2floppy::positionDrive(void)
{
	long fpos;

	if (!tr)
		fpos = pfs->offset +
			((sr - 1) * pfs->byte_p_sector);
	else
		fpos = pfs->offset +
			pfs->byte_p_track0 +
			((sr-1) * pfs->byte_p_sector) +
			((tr-1) * pfs->byte_p_track);
	return fseek(pfs->fp, fpos, SEEK_SET);
}

Byte E2floppy::readByte(unsigned int index)
{

	if (index == pfs->byte_p_sector) {
#ifdef NAFS
		if (pfs->type != NAFS_FORMAT) {
#endif
			if (positionDrive() ||
			   (fread((void *)&sector_buffer, pfs->byte_p_sector, 1, 
			   pfs->fp) != 1)) {
				drq = 0;
				str = 0x10;
				byteCount = 0;
			}
#ifdef NAFS
		} else
			read_nafs_sector();
#endif
	} // if
	return sector_buffer[256 - index];
} // readByte


void E2floppy::writeByte(unsigned int index)
{
	unsigned int error;

	sector_buffer[256 - index] = dr;
	if (index == 1) {
#ifdef NAFS
		if (pfs->type != NAFS_FORMAT) {
#endif
			error =  positionDrive();
			error |= (fwrite((void *)&sector_buffer, pfs->byte_p_sector, 1, 
				pfs->fp) != 1);
#ifdef NAFS
		} else
			write_nafs_sector();
#endif // NAFS
	} // if
} // writeByte


int E2floppy::recordNotFound(void)
{
	if (tr) 
		return (sr == 0 || sr > (pfs->max_sector * 2)); 
	else
		return (sr == 0 || sr > (pfs->max_sector0 * 2)); 
} // recordNotFound

int E2floppy::seekError(Byte new_track)
{
	return (new_track > pfs->max_track);
} // recordNotFound

int E2floppy::driveReady(void)
{
	return pfs->ready;
}  // driveReady

int E2floppy::writeProtect(void)
{
	return pfs->write_protect;
}  // writeProtect

