/*
    mc146818.cc


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/


#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "e2.h"
#include "mc146818.h"
#include "mc6809.h"

Mc146818::Mc146818(Mc6809* x_cpu)
{
	FILE *fp;
	char *home, path[64];

	cpu = x_cpu;
	path[0] = '\0';
	home = getenv("HOME");
	if (home) {
		strcpy(path, home);
	}
	strcat(path, "/.mc146818");
	fp = fopen(path, "rb");
	if (fp) {
		fread(ram, RAM_SIZE, 1, fp);
		fclose(fp);
	}  // if
	A = 0;
	B = 0x06;
	reset();
}

Mc146818::~Mc146818()
{
	FILE *fp;
	char *home, path[64];

	path[0] = '\0';
	home = getenv("HOME");
	if (home) {
		strcpy(path, home);
	}
	strcat(path, "/.mc146818");
	fp = fopen(path, "wb");
	if (fp) {
		fwrite(ram, RAM_SIZE, 1, fp);
		fclose(fp);
	}  // if
}

void Mc146818::reset()
{
	A &= 0x7f;
	B &= 0x87;
	C = 0;
	D = 0x80;
}

Byte Mc146818::read(Word offset)
{
	Byte temp;
	struct tm *lt;
	time_t    time_now;

	time_now = (time_t)time(&time_now);
	lt = localtime(&time_now);
	switch ((offset - RTC_LOW) & 0x3f) {
		case 0x00: return convert(lt->tm_sec); break;
		case 0x01: return 0x00; break;	// alarm not supported
		case 0x02: return convert(lt->tm_min); break;
		case 0x03: return 0x00; break;	// alarm not supported
		case 0x04: return convert_hour(lt->tm_hour); break;
		case 0x05: return 0x00;	break;	// alarm not supported
		case 0x06: return lt->tm_wday + 1; break;
		case 0x07: return convert(lt->tm_mday); break;
		case 0x08: return convert(lt->tm_mon + 1); break;
		case 0x09: return convert(lt->tm_year); break;
		case 0x0a: return A; break;
		case 0x0b: return B; break;
		case 0x0c: temp = C; C = 0x00; return temp; break;
		case 0x0d: return D; break;
		default:   return ram[offset - RTC_LOW - 0x0e]; break;
	}
	return 0; // this case should NEVER happen
}

// writing time not supported

void Mc146818::write(Word offset, Byte val)
{
	switch ((offset - RTC_LOW) & 0x3f) {
		case 0x00: 
		case 0x01: 
		case 0x02: 
		case 0x03: 
		case 0x04: 
		case 0x05: 
		case 0x06: 
		case 0x07: 
		case 0x08: 
		case 0x09: break;
		case 0x0a: A = val & 0x7f; break;
		case 0x0b: B = val & 0x7f; // mask bit SET, makes trouble!!
			   if (B & 0x80)
			   	B &= 0xef; break;
		case 0x0c: break;   // Reg. C is read only
		case 0x0d: break;   // Reg. D is read only
		default:   ram[offset - RTC_LOW - 0x0e] = val; break;
	}
}

void Mc146818::update_1_second(void)
{
	if (!BTST7(B)) {
		BSET4(C);
		if (BTST4(B)) {
			BSET7(C);
			cpu->set_firq();
		}
	}
			

} // update


Byte Mc146818::convert(Byte val)
{
 	if (B & 0x04)
 		return val;
 	else
 		return ((val / 10) << 4)  | (val % 10);
}


Byte Mc146818::convert_hour(Byte val)
{
	switch (B & 0x06) {
		case 0x00:  	//12 hour, BCD
				if (val >= 12)
 					return  0x80 |
 						(((val-12)/10)<<4) | ((val-12)%10);
				else
 					return ((val / 10) << 4)  | (val % 10);
		case 0x02:  	//24 hour, BCD
 				return ((val / 10) << 4)  | (val % 10);
		case 0x04:  	//12 hour, binary
				if (val >= 12)
					return (val - 12) | 0x80;
				else
					return val;
		case 0x06:  	//24 hour, binary
			    	return val;
	}  // switch
	return 0;  // this should NEVER happen
}


