/*
    mc6809.cc

    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This file is based on usim-0.91 which is 
    Copyright (C) 1994 by R. B. Bellis

    woschwot@si.bosch.de
*/



#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <limits.h>
#include <time.h>
#include "machdep.h"
#include "mc6809.h"
#include "memory.h"
#include "mc146818.h"
#include "gui.h"


Mc6809::Mc6809(Memory* x_memory, Da6809* x_disassembler)
	: a(acc.byte.a), b(acc.byte.b), d(acc.d)
{
	disassembler = x_disassembler;
	memory = x_memory;
	init();
	reset();
}

Mc6809::~Mc6809()
{
}

void Mc6809::init(void)
{
	int i, post;
	Byte cycles;

	// all breakpoints are reset
	for (i=0; i < 3; i++)
		reset_bp(i);
	state = RUN;
	new_state = NO_CHANGE;
	for (post = 0; post < 128; post++)
			indexed_cycles[post] = 1;		
	for (post = 128; post < 256; post++)
		switch ((Byte)(post & 0x1f)) {
			case 0x05: case 0x06: case 0x08: case 0x0c:
				indexed_cycles[post] = 1; break;
			case 0x00: case 0x02:
				indexed_cycles[post] = 2; break;
			case 0x01: case 0x03: case 0x14:
				indexed_cycles[post] = 3; break;
			case 0x09: case 0x0b: case 0x18: case 0x15:
			case 0x16: case 0x1c:
				indexed_cycles[post] = 4; break;
			case 0x0d: case 0x1f:
				indexed_cycles[post] = 5; break;
			case 0x11: case 0x13:
				indexed_cycles[post] = 6; break;
			case 0x19: case 0x1b:
				indexed_cycles[post] = 7; break;
			case 0x1d:
				indexed_cycles[post] = 8; break;
			default:
				indexed_cycles[post] = 0; break;
		} // switch

	for (post = 0; post < 256; post++) {
		cycles = 0;
		if (post & 0x01) cycles++;
		if (post & 0x02) cycles++;
		if (post & 0x04) cycles++;
		if (post & 0x08) cycles++;
		if (post & 0x10) cycles += 2;
		if (post & 0x20) cycles += 2;
		if (post & 0x40) cycles += 2;
		if (post & 0x80) cycles += 2;
		psh_pul_cycles[post] = cycles;		
	} // for
} // init

void Mc6809::reset(void)
{
	cycles		= 0;
	next_update	= CYCLES_TO_UPDATE;
	irqs = 0;		// no interrupts yet
	nmi_armed = 0;
	reset_bp(2);	// reset next-breakpoint
	pc = memory->read_word(0xfffe);
	dp = 0x00;		/* Direct page register = 0x00 */
	dp16 = (Word)(dp << 8);
	cc.all = 0x00;		/* Clear all flags */
	cc.bit.i = 1;		/* IRQ disabled */
	cc.bit.f = 1;		/* FIRQ disabled */
}

void Mc6809::get_status(struct s_cpu_status *stat)
{
	time_t time_diff;

	stat->a     = a;
	stat->b     = b;
	stat->cc    = cc.all;
	stat->dp    = dp;
	stat->pc    = pc;
	stat->x     = x;
	stat->y     = y;
	stat->u     = u;
	stat->s     = s;
	stat->state = state;
	stat->cycles = cycles;
	time_diff = time(NULL) - cycle0_time;
	if (time_diff && state == RUN) 
		stat->freq = (cycles - cycle0) / (time_diff * 1000);
	else
		stat->freq = 0;
}  // get_status

int Mc6809::disassemble(unsigned int address, int *pstep,
	char **pb1, char **pb2)
{
	if (disassembler == NULL)
		return 0;
	return disassembler->disassemble(address, pstep, pb1, pb2);
}

void Mc6809::set_new_state(int x_new_state)
{
	new_state = x_new_state;
}

void Mc6809::set_nmi(void)
{
	irqs |= NMI;
} // set_nmi

void Mc6809::set_firq(void)
{
	irqs |= FIRQ;
} // set_firq

void Mc6809::set_irq(void)
{
	irqs |= IRQ;
} // set_irq

t_cycles Mc6809::exec_irqs(void)
{
	if (irqs) {
		if ((irqs & NMI) && !nmi_armed) {
			irqs &= ~NMI; 
			nmi();
		} else if ((irqs & FIRQ) && !cc.bit.f) {
			irqs &= ~FIRQ; 
			firq();
		} else if ((irqs & IRQ) && !cc.bit.i) {
			irqs &= ~IRQ; 
			irq();
		} // else
	}  // if
	return 5; // rounded
}  // exec_irqs

t_cycles Mc6809::execute(void)
{
	/* Select instruction */
	switch ((Byte)(memory->read(pc++))) {
		case 0x00: mode = direct;    neg();  return 6;
		case 0x03: mode = direct;    com();  return 6;
		case 0x04: mode = direct;    lsr();  return 6;
		case 0x06: mode = direct;    ror();  return 6;
		case 0x07: mode = direct;    asr();  return 6;
		case 0x08: mode = direct;    lsl();  return 6;
		case 0x09: mode = direct;    rol();  return 6;
		case 0x0a: mode = direct;    dec();  return 6;
		case 0x0c: mode = direct;    inc();  return 6;
		case 0x0d: mode = direct;    tst();  return 6;
		case 0x0e: mode = direct;    jmp();  return 3;
		case 0x0f: mode = direct;    clr();  return 6;

		case 0x10: return page_10();
		case 0x11: return page_11();
		case 0x12: mode = inherent;  nop();  return 2;
		// sync (0x13) not yet implemented
		case 0x16: lbra(); return 5;
		case 0x17: lbsr(); return 9;
		case 0x19: mode = inherent;  daa();  return 2;
		case 0x1a: mode = immediate; orcc(); return 3;
		case 0x1c: mode = immediate; andcc();return 3;
		case 0x1d: mode = inherent;  sex();  return 2;
		case 0x1e: mode = inherent;  exg();  return 8;
		case 0x1f: mode = inherent;  tfr();  return 6;

		case 0x20: bra();  return 3;
		case 0x21: brn();  return 3;
		case 0x22: bhi();  return 3;
		case 0x23: bls();  return 3;
		case 0x24: bcc();  return 3;
		case 0x25: bcs();  return 3;
		case 0x26: bne();  return 3;
		case 0x27: beq();  return 3;
		case 0x28: bvc();  return 3;
		case 0x29: bvs();  return 3;
		case 0x2a: bpl();  return 3;
		case 0x2b: bmi();  return 3;
		case 0x2c: bge();  return 3;
		case 0x2d: blt();  return 3;
		case 0x2e: bgt();  return 3;
		case 0x2f: ble();  return 3;

		case 0x30: return 4 + leax();
		case 0x31: return 4 + leay();
		case 0x32: return 4 + leas();
		case 0x33: return 4 + leau();
		case 0x34: return 5 + pshs();
		case 0x35: return 5 + puls();
		case 0x36: mode = inherent;  return 5 + pshu();
		case 0x37: mode = inherent;  return 5 + pulu();
		case 0x39: mode = inherent;  rts();  return 5;
		case 0x3a: mode = inherent;  abx();  return 3;
		case 0x3b: mode = inherent;  return rti();
		// cwai (0x3c) not yet implemented
		case 0x3d: mode = inherent;  mul();  return 11;
		case 0x3f: mode = inherent;  swi();  return 19;

		case 0x40: mode = inherent;  nega(); return 2;
		case 0x43: mode = inherent;  coma(); return 2;
		case 0x44: mode = inherent;  lsra(); return 2;
		case 0x46: mode = inherent;  rora(); return 2;	
		case 0x47: mode = inherent;  asra(); return 2;
		case 0x48: mode = inherent;  lsla(); return 2;
		case 0x49: mode = inherent;  rola(); return 2;	
		case 0x4a: mode = inherent;  deca(); return 2;
		case 0x4c: mode = inherent;  inca(); return 2;
		case 0x4d: mode = inherent;  tsta(); return 2;
		case 0x4f: mode = inherent;  clra(); return 2;

		case 0x50: mode = inherent;  negb(); return 2;
		case 0x53: mode = inherent;  comb(); return 2;
		case 0x54: mode = inherent;  lsrb(); return 2;
		case 0x56: mode = inherent;  rorb(); return 2;
		case 0x57: mode = inherent;  asrb(); return 2;
		case 0x58: mode = inherent;  lslb(); return 2;
		case 0x59: mode = inherent;  rolb(); return 2;
		case 0x5a: mode = inherent;  decb(); return 2;
		case 0x5c: mode = inherent;  incb(); return 2;
		case 0x5d: mode = inherent;  tstb(); return 2;
		case 0x5f: mode = inherent;  clrb(); return 2;

		case 0x60: return 6 + neg_i();
		case 0x63: return 6 + com_i();
		case 0x64: return 6 + lsr_i();
		case 0x66: return 6 + ror_i();
		case 0x67: return 6 + asr_i();
		case 0x68: return 6 + lsl_i();
		case 0x69: return 6 + rol_i();
		case 0x6a: return 6 + dec_i();
		case 0x6c: return 6 + inc_i();
		case 0x6d: return 6 + tst_i();
		case 0x6e: return 3 + jmp_i();
		case 0x6f: return 6 + clr_i();

	  	case 0x70: mode = extended;  neg();  return 7;
	 	case 0x73: mode = extended;  com();  return 7;
	 	case 0x74: mode = extended;  lsr();  return 7;
	 	case 0x76: mode = extended;  ror();  return 7;
	 	case 0x77: mode = extended;  asr();  return 7;
	 	case 0x78: mode = extended;  lsl();  return 7;
	 	case 0x79: mode = extended;  rol();  return 3;
	 	case 0x7a: mode = extended;  dec();  return 7;
	 	case 0x7c: mode = extended;  inc();  return 7;
	 	case 0x7d: mode = extended;  tst();  return 7;
	 	case 0x7e: mode = extended;  jmp();  return 4;
	 	case 0x7f: mode = extended;  clr();  return 7;

		case 0x80: mode = immediate; suba(); return 2;
		case 0x81: mode = immediate; cmpa(); return 2;
		case 0x82: mode = immediate; sbca(); return 2;
		case 0x83: mode = immediate; subd(); return 4;
		case 0x84: mode = immediate; anda(); return 2;
		case 0x85: mode = immediate; bita(); return 2;
		case 0x86: mode = immediate; lda();  return 2;
		case 0x88: mode = immediate; eora(); return 2;
		case 0x89: mode = immediate; adca(); return 2;
		case 0x8a: mode = immediate; ora();  return 2;
		case 0x8b: mode = immediate; adda(); return 2;
		case 0x8c: mode = immediate; cmpx(); return 4;
		case 0x8d: bsr();  return 7;
		case 0x8e: mode = immediate; ldx();  return 3;

		case 0x90: mode = direct;    suba(); return 4;
		case 0x91: mode = direct;    cmpa(); return 4;
		case 0x92: mode = direct;    sbca(); return 4;
		case 0x93: mode = direct;    subd(); return 6;
		case 0x94: mode = direct;    anda(); return 4;
		case 0x95: mode = direct;    bita(); return 4;
		case 0x96: mode = direct;    lda();  return 4;
		case 0x97: mode = direct;    sta();  return 4;
		case 0x98: mode = direct;    eora(); return 4;
		case 0x99: mode = direct;    adca(); return 4;
		case 0x9a: mode = direct;    ora();  return 4;
		case 0x9b: mode = direct;    adda(); return 4;
		case 0x9c: mode = direct;    cmpx(); return 6;
		case 0x9d: mode = direct;    jsr();  return 7;
		case 0x9e: mode = direct;    ldx();  return 5;
		case 0x9f: mode = direct;    stx();  return 5;

		case 0xa0: return 4 + suba_i();
		case 0xa1: return 4 + cmpa_i();
		case 0xa2: return 4 + sbca_i();
		case 0xa3: return 6 + subd_i();
		case 0xa4: return 4 + anda_i();
		case 0xa5: return 4 + bita_i();
		case 0xa6: return 4 + lda_i();
		case 0xa7: return 4 + sta_i();
		case 0xa8: return 4 + eora_i();
		case 0xa9: return 4 + adca_i();
		case 0xaa: return 4 + ora_i();
		case 0xab: return 4 + adda_i();
		case 0xac: return 6 + cmpx_i();
		case 0xad: return 7 + jsr_i();
		case 0xae: return 5 + ldx_i();
		case 0xaf: return 5 + stx_i();

		case 0xb0: mode = extended;  suba(); return 5;
		case 0xb1: mode = extended;  cmpa(); return 5;
		case 0xb2: mode = extended;  sbca(); return 5;
		case 0xb3: mode = extended;  subd(); return 7;
		case 0xb4: mode = extended;  anda(); return 5;
		case 0xb5: mode = extended;  bita(); return 5;
		case 0xb6: mode = extended;  lda();  return 5;
	 	case 0xb7: mode = extended;  sta();  return 5;
		case 0xb8: mode = extended;  eora(); return 5;
		case 0xb9: mode = extended;  adca(); return 5;
		case 0xba: mode = extended;  ora();  return 5;
		case 0xbb: mode = extended;  adda(); return 5;
		case 0xbc: mode = extended;  cmpx(); return 7;
	 	case 0xbd: mode = extended;  jsr();  return 8;
		case 0xbe: mode = extended;  ldx();  return 6;
	 	case 0xbf: mode = extended;  stx();  return 6;

		case 0xc0: mode = immediate; subb(); return 2;
		case 0xc1: mode = immediate; cmpb(); return 2;
		case 0xc2: mode = immediate; sbcb(); return 2;
		case 0xc3: mode = immediate; addd(); return 4;
		case 0xc4: mode = immediate; andb(); return 2;
		case 0xc5: mode = immediate; bitb(); return 2;
		case 0xc6: mode = immediate; ldb();  return 2;
		case 0xc8: mode = immediate; eorb(); return 2;
		case 0xc9: mode = immediate; adcb(); return 2;
		case 0xca: mode = immediate; orb();  return 2;
		case 0xcb: mode = immediate; addb(); return 2;
		case 0xcc: mode = immediate; ldd();  return 3;
		case 0xce: mode = immediate; ldu();  return 3;

		case 0xd0: mode = direct;    subb(); return 4;
		case 0xd1: mode = direct;    cmpb(); return 4;
		case 0xd2: mode = direct;    sbcb(); return 4;
		case 0xd3: mode = direct;    addd(); return 6;
		case 0xd4: mode = direct;    andb(); return 4;
		case 0xd5: mode = direct;    bitb(); return 4;
		case 0xd6: mode = direct;    ldb();  return 4;
		case 0xd7: mode = direct;    stb();  return 4;
		case 0xd8: mode = direct;    eorb(); return 4;
		case 0xd9: mode = direct;    adcb(); return 4;
		case 0xda: mode = direct;    orb();  return 4;
		case 0xdb: mode = direct;    addb(); return 4;
		case 0xdc: mode = direct;    ldd();  return 5;
		case 0xdd: mode = direct;    std();  return 5;
		case 0xde: mode = direct;    ldu();  return 5;
		case 0xdf: mode = direct;    stu();  return 5;

		case 0xe0: return 4 + subb_i();
		case 0xe1: return 4 + cmpb_i();
		case 0xe2: return 4 + sbcb_i();
		case 0xe3: return 6 + addd_i();
		case 0xe4: return 4 + andb_i();
		case 0xe5: return 4 + bitb_i();
		case 0xe6: return 4 + ldb_i();
		case 0xe7: return 4 + stb_i();
		case 0xe8: return 4 + eorb_i();
		case 0xe9: return 4 + adcb_i();
		case 0xea: return 4 + orb_i();
		case 0xeb: return 4 + addb_i();
		case 0xec: return 5 + ldd_i();
		case 0xed: return 5 + std_i();
		case 0xee: return 5 + ldu_i();
		case 0xef: return 5 + stu_i();

		case 0xf0: mode = extended;  subb(); return 5;
		case 0xf1: mode = extended;  cmpb(); return 5;
		case 0xf2: mode = extended;  sbcb(); return 5;
		case 0xf3: mode = extended;  addd(); return 7;
		case 0xf4: mode = extended;  andb(); return 5;
		case 0xf5: mode = extended;  bitb(); return 5;
		case 0xf6: mode = extended;  ldb();  return 5;
	 	case 0xf7: mode = extended;  stb();  return 5;
		case 0xf8: mode = extended;  eorb(); return 5;
		case 0xf9: mode = extended;  adcb(); return 5;
		case 0xfa: mode = extended;  orb();  return 5;
		case 0xfb: mode = extended;  addb(); return 5;
		case 0xfc: mode = extended;  ldd();  return 6;
	 	case 0xfd: mode = extended;  std();  return 6;
		case 0xfe: mode = extended;  ldu();  return 6;
	 	case 0xff: mode = extended;  stu();  return 6;

		default:   invalid("instruction"); return 0;
	}
}

t_cycles Mc6809::page_10(void)
{
	/* Select instruction */
	switch ((Byte)(memory->read(pc++))) {
		case 0x21: lbrn(); return 5;
		case 0x22: return 5 + lbhi();
		case 0x23: return 5 + lbls();
		case 0x24: return 5 + lbcc();
		case 0x25: return 5 + lbcs();
		case 0x26: return 5 + lbne();
		case 0x27: return 5 + lbeq();
		case 0x28: return 5 + lbvc();
		case 0x29: return 5 + lbvs();
		case 0x2a: return 5 + lbpl();
		case 0x2b: return 5 + lbmi();
		case 0x2c: return 5 + lbge();
		case 0x2d: return 5 + lblt();
		case 0x2e: return 5 + lbgt();
		case 0x2f: return 5 + lble();

		case 0x3f: mode = inherent; swi2();  return 20;

		case 0x83: mode = immediate; cmpd(); return 5;
		case 0x8c: mode = immediate; cmpy(); return 5;
		case 0x8e: mode = immediate; ldy();  return 4;

		case 0x93: mode = direct;    cmpd(); return 7;
		case 0x9c: mode = direct;    cmpy(); return 7;
		case 0x9e: mode = direct;    ldy();  return 6;
		case 0x9f: mode = direct;    sty();  return 6;

		case 0xa3: return 7 + cmpd_i();
		case 0xac: return 7 + cmpy_i();
		case 0xae: return 6 + ldy_i();
		case 0xaf: return 6 + sty_i();
		
		case 0xb3: mode = extended;  cmpd(); return 8;
		case 0xbc: mode = extended;  cmpy(); return 8;
		case 0xbe: mode = extended;  ldy();  return 7;
	 	case 0xbf: mode = extended;  sty();  return 7;

		case 0xce: mode = immediate; lds();  return 4;

		case 0xde: mode = direct;    lds();  return 6;
		case 0xdf: mode = direct;    sts();  return 6;

		case 0xee: return 6 + lds_i();
		case 0xef: return 6 + sts_i();

		case 0xfe: mode = extended;  lds();  return 7;
	 	case 0xff: mode = extended;  sts();  return 7;

		default:   --pc; invalid("instruction"); return 0;
	}
} // page_10


t_cycles Mc6809::page_11(void)
{
	/* Select instruction */
	switch ((Byte)(memory->read(pc++))) {
		case 0x3f: mode = inherent;  swi3(); return 20;
		case 0x83: mode = immediate; cmpu(); return 5;
		case 0x8c: mode = immediate; cmps(); return 5;
		case 0x93: mode = direct;    cmpu(); return 7;
		case 0x9c: mode = direct;    cmps(); return 7;
		case 0xa3: return 7 + cmpu_i();
		case 0xac: return 7 + cmps_i();
		case 0xb3: mode = extended;  cmpu(); return 8;
		case 0xbc: mode = extended;  cmps(); return 8;
		default:   --pc; invalid("instruction"); return 0;
	}
} // page_11




Word& Mc6809::refreg(Byte post)
{
// Sw: enhanced
	switch ((Byte)(post & 0x60)) {
		case 0x00 : return x;
		case 0x20 : return y;
		case 0x40 : return u;
		case 0x60 : return s;
		case 0x80 : return s; // one more case for performance
		default   : return s; // should never happen
	} // switch
} // refreg


Byte Mc6809::fetch_operand(void)
{
// Sw: enhanced
	Byte		ret = 0;
	Word		addr;

	switch (mode) {
		case immediate :
		case relative  : return memory->read(pc++);  //is identical to case relative
		case extended  : addr = memory->read_word(pc);
				 pc += 2;
			         return memory->read(addr);
		case direct    : addr = dp16 | memory->read(pc++);
		                 return memory->read(addr);
		default        : invalid("addressing mode");
				 return ret;
	}  // switch
}  // fetch_operand


Word Mc6809::fetch_word_operand(void)
{
	Word addr = 0;

	switch(mode) {
		case immediate :
		case relative  : addr = memory->read_word(pc);
				 pc += 2;
				 return addr;
		case extended  : addr = memory->read_word(pc);
				 pc += 2;
	 			 return memory->read_word(addr);
		case direct    : addr = dp16 | memory->read(pc++);
				 return memory->read_word(addr);
		default        : invalid("addressing mode");
				 return addr;
	} // switch
}  // fetch_word_operand

Word Mc6809::fetch_effective_address(void)
{
	Word		addr = 0;

	if (mode == extended) {
		addr = memory->read_word(pc);
		pc += 2;
		return addr;
	} else 
	if (mode == direct)
		return dp16 | memory->read(pc++);
	else
		invalid("addressing mode");
	return addr;
} // fetch_effective_address

Word Mc6809::fetch_indexed_address(t_cycles *c)
{
	Word		addr;
	Byte		post;

	post = memory->read(pc++);
	if ((post & 0x8e) == 0x82)
		do_predecrement(post);
	addr = do_effective_address(post);
	if ((post & 0x8e) == 0x80)
		do_postincrement(post);
	*c = indexed_cycles[post];
	return addr;
} // fetch_indexed_address

Word Mc6809::fetch_indexed_word(t_cycles *c)
{
	Word		addr;
	Byte		post;

	post = memory->read(pc++);
	if ((post & 0x8e) == 0x82)
		do_predecrement(post);
	addr = do_effective_address(post);
	if ((post & 0x8e) == 0x80)
		do_postincrement(post);
	*c = indexed_cycles[post];
	return memory->read_word(addr);
}

Byte Mc6809::fetch_indexed(t_cycles *c)
{
	Word		addr;
	Byte		post;
	Byte		ret;

	post = memory->read(pc++);
	if ((post & 0x8e) == 0x82)
		do_predecrement(post);
        addr = do_effective_address(post);
        ret = memory->read(addr);
	if ((post & 0x8e) == 0x80)
        	do_postincrement(post);
	*c = indexed_cycles[post];
        return ret;
}

Word Mc6809::do_effective_address(Byte post)
{
	Word addr = 0;
	Word offset;

	if ((post & 0x80) == 0x00) {
		offset = post & 0x1f;
		if (offset & 0x10)
			offset |= 0xffe0;
		addr = refreg(post) + offset;
	} else {
		switch ((Byte)(post & 0x7f)) {
			case 0x00: case 0x01: case 0x02: case 0x03: // X
			case 0x04: case 0x14: case 0x11: case 0x13:
				addr = x;
				break;
			case 0x20: case 0x21: case 0x22: case 0x23: // Y
			case 0x24: case 0x31: case 0x33: case 0x34:
				addr = y;
				break;
			case 0x40: case 0x41: case 0x42: case 0x43: // U
			case 0x44: case 0x51: case 0x53: case 0x54:
				addr = u;
				break;
			case 0x60: case 0x61: case 0x62: case 0x63: // S
			case 0x64: case 0x71: case 0x73: case 0x74:
				addr = s;
				break;
			case 0x05: case 0x15:
				addr = extend8(b) + x;
				break;
			case 0x25: case 0x35:
				addr = extend8(b) + y;
				break;
			case 0x45: case 0x55:
				addr = extend8(b) + u;
				break;
			case 0x65: case 0x75:
				addr = extend8(b) + s;
				break;
			case 0x06: case 0x16:
				addr = extend8(a) + x;
				break;
			case 0x26: case 0x36:
				addr = extend8(a) + y;
				break;
			case 0x46: case 0x56:
				addr = extend8(a) + u;
				break;
			case 0x66: case 0x76:
				addr = extend8(a) + s;
				break;
			case 0x08: case 0x18:
				addr = x + extend8(memory->read(pc++));
				break;
			case 0x28: case 0x38:
				addr = y + extend8(memory->read(pc++));
				break;
			case 0x48: case 0x58:
				addr = u + extend8(memory->read(pc++));
				break;
			case 0x68: case 0x78:
				addr = s + extend8(memory->read(pc++));
				break;
			case 0x09: case 0x19:
				addr = x + memory->read_word(pc);
				pc += 2;
				break;
			case 0x29: case 0x39:
				addr = y + memory->read_word(pc);
				pc += 2;
				break;
			case 0x49: case 0x59:
				addr = u + memory->read_word(pc);
				pc += 2;
				break;
			case 0x69: case 0x79:
				addr = s + memory->read_word(pc);
				pc += 2;
				break;
			case 0x0b: case 0x1b:
				addr = d + x;
				break;
			case 0x2b: case 0x3b:
				addr = d + y;
				break;
			case 0x4b: case 0x5b:
				addr = d + u;
				break;
			case 0x6b: case 0x7b:
				addr = d + s;
				break;
			case 0x0c: case 0x1c: case 0x2c: case 0x3c:
			case 0x4c: case 0x5c: case 0x6c: case 0x7c:
				addr = extend8(memory->read(pc++));
				addr += pc;
				break;
			case 0x0d: case 0x1d: case 0x2d: case 0x3d:
			case 0x4d: case 0x5d: case 0x6d: case 0x7d:
				addr = memory->read_word(pc);
				pc += 2;
				addr += pc;
				break;
			case 0x1f: case 0x3f: case 0x5f: case 0x7f:
				addr = memory->read_word(pc);
				pc += 2;
				break;
			default: --pc;
				invalid("indirect addressing postbyte");
				break;
		}

		/* Do extra indirection */
		if (post & 0x10) {
			addr = memory->read_word(addr);
		}
	}

	return addr;
}

void Mc6809::do_postincrement(Byte post)
{
	switch (post) {
		case 0x80: x++; break;
		case 0xa0: y++; break;
		case 0xc0: u++; break;
		case 0xe0: s++; break;
		case 0x90: case 0xb0: case 0xd0: case 0xf0:
			--pc;
			invalid("postincrement");
			break;
		case 0x91:
		case 0x81: x += 2; break;
		case 0xb1:
		case 0xa1: y += 2; break;
		case 0xd1:
		case 0xc1: u += 2; break;
		case 0xf1:
		case 0xe1: s += 2; break;
	} // switch
} // do_postincrement

void Mc6809::do_predecrement(Byte post)
{
	switch (post) {
		case 0x82: x--; break;
		case 0xa2: y--; break;
		case 0xc2: u--; break;
		case 0xe2: s--; break;
		case 0x92: case 0xb2: case 0xd2: case 0xf2:
			--pc;
			invalid("predecrement");
			break;
			case 0x91:
		case 0x83: x -= 2; break;
		case 0xb3:
		case 0xa3: y -= 2; break;
		case 0xd3:
		case 0xc3: u -= 2; break;
		case 0xf3:
		case 0xe3: s -= 2; break;
	} // switch
} // do_predecrement


void Mc6809::invalid(const char *msg)
{
	sprintf((char *)&err_msg, "\
invalid %s\n\
pc=%04x opcode=%02x\n\
processor stopped\n", msg ? msg : "", pc-1, memory->read(pc-1));
#ifndef XLIB
	popup_message(err_msg);
#else
	fprintf(stderr, err_msg);
#endif
	irqs |= INVALID;
}

#ifndef XLIB
void Mc6809::popup_message(char *pmsg)
{
	if (gui)
		gui->popup_message(pmsg);
	else
		fprintf(stderr, pmsg);
}
#endif

//----------------------------------------------------------------------------
// Processor loading routines
//----------------------------------------------------------------------------
static Byte fread_byte(FILE *fp)
{
	char			str[3];
	long			l;

	str[0] = fgetc(fp);
	str[1] = fgetc(fp);
	str[2] = '\0';

	l = strtol(str, NULL, 16);
	return (Byte)(l & 0xff);
}

static Word fread_word(FILE *fp)
{
	Word		ret;

	ret = fread_byte(fp);
	ret <<= 8;
	ret |= fread_byte(fp);

	return ret;
}

void Mc6809::load_intelhex(FILE *fp)
{
	int		done = 0;

	while (!done) {
		Byte		n, t;
		Word		addr;
		Byte		b;

		(void)fgetc(fp);
		n = fread_byte(fp);
		addr = fread_word(fp);
		t = fread_byte(fp);
		if (t == 0x00) {
			while (n--) {
				b = fread_byte(fp);
				memory->write_rom(addr++, b);
			}
		} else if (t == 0x01) {
			pc = addr;
			done = 1;
		}
		// Read and discard checksum byte
		(void)fread_byte(fp);
		if (fgetc(fp) == '\r') (void)fgetc(fp);
	}
}

void Mc6809::load_motorola_srec(FILE *fp)
{
	int		done = 0;

	while (!done) {
		Byte		n, t;
		Word		addr;
		Byte		b;

		(void)fgetc(fp); /* read 'S' */
		t = fgetc(fp);   /* read type of line */
		n = fread_byte(fp);
		switch (t) {
		case '0': n -= 1;
			   while (n--) {
				b = fread_byte(fp);
			   }; 
			   break;
		case '1': n -= 3;
			   addr = fread_word(fp);
		           while (n--) {
				b = fread_byte(fp);
				memory->write_rom(addr++, b);
			   }; 
			   break;
		case '9': addr = fread_word(fp);
			   pc = addr;
			   done = 1;
			   break;
		}
		// Read and discard checksum byte
		(void)fread_byte(fp);
		if (fgetc(fp) == '\r') (void)fgetc(fp);
	}
	fclose(fp);
}

void Mc6809::load_hexfile(const char *filename)
{
	FILE		*fp;
	int		ch;
	char		*phome, *ppath;

	// try to open in working directory
	fp = fopen(filename, "r");
	if (fp == NULL) {
		// try to open in the homedirectory 
		phome = getenv("HOME");
		if (phome != NULL) {
			ppath = (char *)calloc(1, strlen(phome) +
				strlen(filename) + 2);
			if (ppath != NULL) {
				strcpy(ppath, phome);
				strcat(ppath, "/");
				strcat(ppath, filename);
				fp = fopen(ppath, "r");
				free(ppath);
			} // if
		} // if
		if (fp == NULL) {
			perror(filename);
			exit(EXIT_FAILURE);
		} // if
	} // if
	ch = fgetc(fp);
	ungetc(ch, fp);
	if (ch == ':')
		load_intelhex(fp);
	else if (toupper(ch) == 'S')
		load_motorola_srec(fp);
	else {
		perror("fileformat unknown");	
		exit(EXIT_FAILURE);
	}
}  // load_hexfile


void Mc6809::set_bp(int which, Word address)
{
	bp[which] = address;
	bp_set = 1;
}

unsigned int Mc6809::get_bp(int which)
{
	return bp[which];
}

int Mc6809::is_bp_set(int which)
{
	return bp[which] < 0x10000;
}

void Mc6809::reset_bp(int which)
{
	bp[which] = 0x10000;
	if (bp[0] > 0xffff && bp[1] > 0xffff && bp[2] > 0xffff)
		bp_set = 0;
}

void Mc6809::set_gui(Gui* x_gui)
{
	gui = x_gui;
}  // set_gui


void Mc6809::update_1_second(unsigned int state)
{
	if (gui)
		gui->update_1_second(state);	// update status display
	memory->rtc->update_1_second();
} // update_1_second


unsigned int Mc6809::run(void)
{
	next_update = cycles + CYCLES_TO_UPDATE;
	cycle0_time = time(NULL);
	cycle0 = cycles;
	while (1) {
		if (irqs) {
			// check if invalid instr. occured
			if (irqs & INVALID) {
				irqs &= ~INVALID;
				set_new_state(STOP);
				return new_state;
			}
			exec_irqs();
		}
		cycles += execute();
		if (bp_set && (pc == bp[0] || pc == bp[1] || pc == bp[2])) {
			if (pc == bp[2]) {
				bp[2] = 0x10000L;
				if (bp[0] > 0xffff && bp[1] > 0xffff && bp[2] > 0xffff)
					bp_set = 0;
			} // if
			return STOP;
		} // if
		if ((cycles >= next_update)) {
			if (update_flag) {
				update_1_second(state);
				update_flag = 0;
			}
			next_update = cycles + CYCLES_TO_UPDATE;
			if (next_update >= ULONG_MAX - 20) {
				next_update = 0;
				cycle0_time = time(NULL);
				cycle0 = 0;
			}
			if (gui)
				gui->update();
			if (new_state != NO_CHANGE &&
				new_state != STEP &&
				new_state != NEXT)
				return new_state;
		} // if
	} // while
} // run


unsigned int Mc6809::step(unsigned int previous_state)
{
	cycles += execute();
	return previous_state;
}

unsigned int Mc6809::do_reset(unsigned int previous_state)
{
	reset();
	memory->reset();
	return previous_state;
}

unsigned int Mc6809::stop(void)
{
	char *pa, *pb;
	int step;

	new_state = NO_CHANGE;
	while (1) {
		if (gui)
			gui->update();	// update video display
		if (new_state != NO_CHANGE) {
			if (new_state == NEXT && disassembler) {
				bp[2] =
			 	   pc + disassembler->disassemble(
					   (unsigned int)pc, &step,
					   &pa, &pb);
				new_state = step ? STEP : RUN;
				bp_set = step ? 0 : 1;
			}
			return new_state;
		}
	} // while
}	


int Mc6809::statemachine(void)
{
	unsigned int prev_state; 

	prev_state = state;
	while(1) {
		if (gui)
			gui->update_1_second(NO_CHANGE);
		new_state = NO_CHANGE;
		switch (state) {
			case RUN:   prev_state = state; state = run(); break;
			case STOP:  prev_state = state; state = stop(); break;
			case RESET: state = do_reset(prev_state); break;
			case STEP:  state = step(prev_state); break;
			case EXIT:  exit(0);
		} // switch
	} // while
} // statemachine

