/*
    memory.cc


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/



#include "memory.h"
#include "pia1.h"
#include "pia2.h"
#include "e2floppy.h"
#include "mc146818.h"
#include "e2video.h"

Memory::Memory()
{
	int i;

	for (i=0; i < YBLOCKS; i++)
		changed[i] = 0;
	init();
} // constructor

Memory::~Memory()
{
} // destructor

// memory must be initialized AFTER all memory mapped I/O is created
void Memory::init(void)
{
	int i;

	for (i = 0; i < 16; i++)
		ppage[i] = NULL;
	for (i = 0; i <= (~GENIO_MASK & 0xffff); i++)
		if (((GENIO_BASE+i) & FDCB_MASK) == FDCB_BASE)
			read_tab[i] = FDC_READSTATUS;
		else
		if (((GENIO_BASE+i) & FDCA_MASK) == FDCA_BASE)
			read_tab[i] = FDC_READ;
		else
		if (((GENIO_BASE+i) & PIA1_MASK) == PIA1_BASE)
			read_tab[i] = PIA1_READ;
		else
		if (((GENIO_BASE+i) & PIA2_MASK) == PIA2_BASE)
			read_tab[i] = PIA2_READ;
		else
		if ((GENIO_BASE+i) >= RTC_LOW && ((GENIO_BASE+i) <= RTC_HIGH))
			read_tab[i] = RTC_READ;
		else
		if (((GENIO_BASE+i) & MMU_MASK) == MMU_BASE)
			read_tab[i] = READ_UNSELECTED;
		else
		if (((GENIO_BASE+i) & VICO_MASK) == VICO_BASE)
			read_tab[i] = READ_UNSELECTED;
		else
			read_tab[i] = READ_DEFAULT;

	for (i = 0; i <= (~GENIO_MASK & 0xffff); i++)
		if (((GENIO_BASE+i) & FDCA_MASK) == FDCA_BASE)
			write_tab[i] = FDC_WRITE;
		else
		if (((GENIO_BASE+i) & FDCB_MASK) == FDCB_BASE)
			write_tab[i] = FDC_WRITEDRISEL;
		else
		if (((GENIO_BASE+i) & MMU_MASK) == MMU_BASE)
			write_tab[i] = WRITE_MMU;
		else
		if (((GENIO_BASE+i) & PIA1_MASK) == PIA1_BASE)
			write_tab[i] = PIA1_WRITE;
		else
		if (((GENIO_BASE+i) & PIA2_MASK) == PIA2_BASE)
			write_tab[i] = PIA2_WRITE;
		else
		if (((GENIO_BASE+i) & VICO_MASK) == VICO_BASE)
			write_tab[i] = VIDEO_WRITE_CO;
		else
		if ((GENIO_BASE+i) >= RTC_LOW && ((GENIO_BASE+i) <= RTC_HIGH))
			write_tab[i] = RTC_WRITE;
		else
			write_tab[i] = NO_WRITE;
} // init


void Memory::initialize_byte_conv_table(int lsb_first)
{
	Byte new_val;
	int i, bitpos, new_bitpos;

	for (i=0; i < 256; i++) {
		if (lsb_first) {
			new_val = 0;
			new_bitpos = 0x80;
			for (bitpos=1; bitpos <= 0x80; bitpos <<= 1) {
				if (i & bitpos)
					new_val |= new_bitpos;
				new_bitpos >>= 1;
			} // for
		} else
			new_val = i;
		byte_conversion_table[i] = new_val;
	}  // for 
}


void Memory::reset(void)
{
	pia1->reset();
	pia2->reset();
	fdc->reset();
	rtc->reset();
} // reset



Byte Memory::read(Word addr)
{
	Byte *p;

	// read from memory mapped I/O
	if ((addr & GENIO_MASK) == GENIO_BASE)
		switch (read_tab[addr - GENIO_BASE]) {
			case FDC_READSTATUS:	return fdc->readstatus(addr);
			case FDC_READ:		return fdc->read(addr);
			case PIA1_READ:		return pia1->read(addr);
			case PIA2_READ:		return pia2->read(addr);
			case RTC_READ:		return rtc->read(addr);
			case READ_UNSELECTED:	return read_unselected(addr);
		} // switch
	// otherwise read from memory
	p = ppage[addr >> 12];
	if (p)
		return byte_conversion_table[(Byte)*(p + (addr & 0x3fff))];
	else
		return memory[addr]; 
} // read


void Memory::write(Word addr, Byte x)
{
	Word vr_offset;  // offset into video ram (modulo 0x4000)
	Byte *p;

	if (addr >= ROM_BASE)
		switch (write_tab[addr - GENIO_BASE]) {
			case FDC_WRITE:		fdc->write(addr, x); break;
			case FDC_WRITEDRISEL:	fdc->writedrisel(addr, x); break;
			case WRITE_MMU:		write_mmu(addr, x); break;
			case PIA1_WRITE:	pia1->write(addr, x); break;
			case PIA2_WRITE:	pia2->write(addr, x); break;
			case VIDEO_WRITE_CO:	video->write_co(addr, x); break;
			case RTC_WRITE:		rtc->write(addr, x); break;
		} // switch
	else {
		p = ppage[addr >> 12];
		if (p) {
			vr_offset = addr & 0x3fff;
			changed[vr_offset / YBLOCK_SIZE] = 1;
			*(p + vr_offset) = byte_conversion_table[x];
		} else
			memory[addr] = x;
	} // else
} // write


void Memory::write_rom(Word offset, Byte val)
{
	memory[offset] = val;
} // write_rom


void Memory::write_mmu(Word offset, Byte val)
{
	int index;
	Byte *p;

	index = offset - MMU_BASE;
	switch ((Byte)(val & 0x0f)) {
		case 0x04: p = &video_ram[VIDEORAM_SIZE * 3]; break;
		case 0x05: p = &video_ram[VIDEORAM_SIZE * 4]; break;
		case 0x06: p = &video_ram[VIDEORAM_SIZE * 5]; break;
		case 0x0c: p = &video_ram[VIDEORAM_SIZE * 0]; break;
		case 0x0d: p = &video_ram[VIDEORAM_SIZE * 1]; break;
		case 0x0e: p = &video_ram[VIDEORAM_SIZE * 2]; break;
		case 0x00: p = &video_ram[VIDEORAM_SIZE * 9]; break;
		case 0x01: p = &video_ram[VIDEORAM_SIZE * 10]; break;
		case 0x02: p = &video_ram[VIDEORAM_SIZE * 11]; break;
		case 0x08: p = &video_ram[VIDEORAM_SIZE * 6]; break;
		case 0x09: p = &video_ram[VIDEORAM_SIZE * 7]; break;
		case 0x0a: p = &video_ram[VIDEORAM_SIZE * 8]; break;
		default  : p = NULL; break;
	} // switch
	if (index < 12)
		ppage[index] = p; 
} // write_mmu


// only for better compatibility (three-state bus usually all one's
// if nothing selected

Byte Memory::read_unselected(Word offset)
{
	return 0xff;
}

void Memory::write_word(Word offset, Word val)
{
	write(offset++, (Byte)(val >> 8));
	write(offset, (Byte)val);
} // write_word

Word Memory::read_word(Word offset)
{
	Word tmp;

	tmp = (read(offset++) << 8) | read(offset);
	return tmp;
} // read_word

